@@ -1,4 +1,5 @@
 #include <unistd.h>
+#include <linux/netlink.h>
 #include <sys/socket.h>
 #include <sys/types.h>
 #include <sys/eventfd.h>
@@ -7,11 +8,13 @@
 #include <sys/signalfd.h>
 #include <sys/ptrace.h>
 #include <sys/wait.h>
+#include <sys/socket.h>
 #include <fcntl.h>
 #include <signal.h>
 #include <linux/if.h>
 #include <sys/ioctl.h>
 #include <termios.h>
+#include <sys/mman.h>
 
 #include "proc_parse.h"
 #include "sockets.h"
@@ -94,8 +97,16 @@ static int check_map_files(void)
 static int check_sock_diag(void)
 {
 	int ret;
+	struct ns_id ns;
 
-	ret = collect_sockets(0);
+	ns.pid = 0;
+	ns.net.nlsk = socket(PF_NETLINK, SOCK_RAW, NETLINK_SOCK_DIAG);
+	if (ns.net.nlsk < 0) {
+		pr_perror("Can't make diag socket for check");
+		return -1;
+	}
+
+	ret = collect_sockets(&ns);
 	if (!ret)
 		return 0;
 
@@ -155,6 +166,7 @@ static int check_prctl(void)
 {
 	unsigned long user_auxv = 0;
 	unsigned int *tid_addr;
+	unsigned int size = 0;
 	int ret;
 
 	ret = sys_prctl(PR_GET_TID_ADDRESS, (unsigned long)&tid_addr, 0, 0, 0);
@@ -163,25 +175,38 @@ static int check_prctl(void)
 		return -1;
 	}
 
-	ret = sys_prctl(PR_SET_MM, PR_SET_MM_BRK, sys_brk(0), 0, 0);
+	/*
+	 * Either new or old interface must be supported in the kernel.
+	 */
+	ret = sys_prctl(PR_SET_MM, PR_SET_MM_MAP_SIZE, (unsigned long)&size, 0, 0);
 	if (ret) {
-		if (ret == -EPERM)
-			pr_msg("prctl: One needs CAP_SYS_RESOURCE capability to perform testing\n");
-		else
-			pr_msg("prctl: PR_SET_MM is not supported\n");
-		return -1;
-	}
+		if (!opts.check_ms_kernel) {
+			pr_msg("prctl: PR_SET_MM_MAP is not supported, which "
+			       "is required for restoring user namespaces\n");
+			return -1;
+		} else
+			pr_warn("Skipping unssuported PR_SET_MM_MAP\n");
 
-	ret = sys_prctl(PR_SET_MM, PR_SET_MM_EXE_FILE, -1, 0, 0);
-	if (ret != -EBADF) {
-		pr_msg("prctl: PR_SET_MM_EXE_FILE is not supported (%d)\n", ret);
-		return -1;
-	}
+		ret = sys_prctl(PR_SET_MM, PR_SET_MM_BRK, sys_brk(0), 0, 0);
+		if (ret) {
+			if (ret == -EPERM)
+				pr_msg("prctl: One needs CAP_SYS_RESOURCE capability to perform testing\n");
+			else
+				pr_msg("prctl: PR_SET_MM is not supported\n");
+			return -1;
+		}
 
-	ret = sys_prctl(PR_SET_MM, PR_SET_MM_AUXV, (long)&user_auxv, sizeof(user_auxv), 0);
-	if (ret) {
-		pr_msg("prctl: PR_SET_MM_AUXV is not supported\n");
-		return -1;
+		ret = sys_prctl(PR_SET_MM, PR_SET_MM_EXE_FILE, -1, 0, 0);
+		if (ret != -EBADF) {
+			pr_msg("prctl: PR_SET_MM_EXE_FILE is not supported (%d)\n", ret);
+			return -1;
+		}
+
+		ret = sys_prctl(PR_SET_MM, PR_SET_MM_AUXV, (long)&user_auxv, sizeof(user_auxv), 0);
+		if (ret) {
+			pr_msg("prctl: PR_SET_MM_AUXV is not supported\n");
+			return -1;
+		}
 	}
 
 	return 0;
@@ -189,11 +214,22 @@ static int check_prctl(void)
 
 static int check_fcntl(void)
 {
-	/*
-	 * FIXME Add test for F_GETOWNER_UIDS once
-	 * it's merged into mainline and kernel part
-	 * settle down.
-	 */
+	u32 v[2];
+	int fd;
+
+	fd = open("/proc/self/comm", O_RDONLY);
+	if (fd < 0) {
+		pr_perror("Can't open self comm file");
+		return -1;
+	}
+
+	if (fcntl(fd, F_GETOWNER_UIDS, (long)v)) {
+		pr_perror("Can'r fetch file owner UIDs");
+		close(fd);
+		return -1;
+	}
+
+	close(fd);
 	return 0;
 }
 
@@ -299,7 +335,8 @@ static int check_fdinfo_signalfd(void)
 
 static int check_one_epoll(union fdinfo_entries *e, void *arg)
 {
-	*(int *)arg = e->epl.tfd;
+	*(int *)arg = e->epl.e.tfd;
+	free_event_poll_entry(e);
 	return 0;
 }
 
@@ -353,7 +390,8 @@ pipe_err:
 
 static int check_one_inotify(union fdinfo_entries *e, void *arg)
 {
-	*(int *)arg = e->ify.wd;
+	*(int *)arg = e->ify.e.wd;
+	free_inotify_wd_entry(e);
 	return 0;
 }
 
@@ -477,7 +515,7 @@ static int check_ipc(void)
 	return -1;
 }
 
-int check_sigqueuinfo()
+static int check_sigqueuinfo()
 {
 	siginfo_t info = { .si_code = 1 };
 
@@ -491,7 +529,7 @@ int check_sigqueuinfo()
 	return 0;
 }
 
-int check_ptrace_peeksiginfo()
+static int check_ptrace_peeksiginfo()
 {
 	struct ptrace_peeksiginfo_args arg;
 	siginfo_t siginfo;
@@ -507,8 +545,11 @@ int check_ptrace_peeksiginfo()
 		exit(1);
 	}
 
-	if (ptrace(PTRACE_ATTACH, pid, NULL, NULL) == -1)
-		return -1;
+	if (ptrace(PTRACE_ATTACH, pid, NULL, NULL) == -1) {
+		pr_perror("Unable to ptrace the child");
+		ret = -1;
+		goto out;
+	}
 
 	waitpid(pid, NULL, 0);
 
@@ -526,8 +567,8 @@ int check_ptrace_peeksiginfo()
 		ret = -1;
 	}
 
-	ptrace(PTRACE_KILL, pid, NULL, NULL);
-
+out:
+	kill(pid, SIGKILL);
 	return ret;
 }
 
@@ -536,7 +577,7 @@ static int check_mem_dirty_track(void)
 	if (kerndat_get_dirty_track() < 0)
 		return -1;
 
-	if (!kerndat_has_dirty_track)
+	if (!kdat.has_dirty_track)
 		pr_warn("Dirty tracking is OFF. Memory snapshot will not work.\n");
 	return 0;
 }
@@ -553,13 +594,90 @@ static int check_posix_timers(void)
 	return -1;
 }
 
+static unsigned long get_ring_len(unsigned long addr)
+{
+	FILE *maps;
+	char buf[256];
+
+	maps = fopen("/proc/self/maps", "r");
+	if (!maps) {
+		pr_perror("No maps proc file");
+		return 0;
+	}
+
+	while (fgets(buf, sizeof(buf), maps)) {
+		unsigned long start, end;
+		int r, tail;
+
+		r = sscanf(buf, "%lx-%lx %*s %*s %*s %*s %n\n", &start, &end, &tail);
+		if (r != 2) {
+			fclose(maps);
+			pr_err("Bad maps format %d.%d (%s)\n", r, tail, buf + tail);
+			return 0;
+		}
+
+		if (start == addr) {
+			fclose(maps);
+			if (strcmp(buf + tail, "/[aio] (deleted)\n"))
+				goto notfound;
+
+			return end - start;
+		}
+	}
+
+	fclose(maps);
+notfound:
+	pr_err("No AIO ring at expected location\n");
+	return 0;
+}
+
+static int check_aio_remap(void)
+{
+	aio_context_t ctx = 0;
+	unsigned long len;
+	void *naddr;
+	int r;
+
+	if (sys_io_setup(16, &ctx) < 0) {
+		pr_err("No AIO syscall");
+		return -1;
+	}
+
+	len = get_ring_len((unsigned long) ctx);
+	if (!len)
+		return -1;
+
+	naddr = mmap(NULL, len, PROT_READ | PROT_WRITE, MAP_ANON | MAP_PRIVATE, 0, 0);
+	if (naddr == MAP_FAILED) {
+		pr_perror("Can't find place for new AIO ring\n");
+		return -1;
+	}
+
+	if (mremap((void *)ctx, len, len, MREMAP_FIXED | MREMAP_MAYMOVE, naddr) == MAP_FAILED) {
+		pr_perror("Can't remap AIO ring\n");
+		return -1;
+	}
+
+	ctx = (aio_context_t)naddr;
+	r = sys_io_getevents(ctx, 0, 1, NULL, NULL);
+	if (r < 0) {
+		if (!opts.check_ms_kernel) {
+			pr_err("AIO remap doesn't work properly\n");
+			return -1;
+		} else
+			pr_warn("Skipping unsupported AIO remap\n");
+	}
+
+	return 0;
+}
+
+static int (*chk_feature)(void);
+
 int cr_check(void)
 {
 	struct ns_id ns = { .pid = getpid(), .nd = &mnt_ns_desc };
 	int ret = 0;
 
-	log_set_loglevel(LOG_WARN);
-
 	if (!is_root_user())
 		return -1;
 
@@ -578,6 +696,11 @@ int cr_check(void)
 	if (mntinfo == NULL)
 		return -1;
 
+	if (chk_feature) {
+		ret = chk_feature();
+		goto out;
+	}
+
 	ret |= check_map_files();
 	ret |= check_sock_diag();
 	ret |= check_ns_last_pid();
@@ -596,12 +719,65 @@ int cr_check(void)
 	ret |= check_ptrace_peeksiginfo();
 	ret |= check_mem_dirty_track();
 	ret |= check_posix_timers();
-	ret |= check_tun();
+	ret |= check_tun_cr(0);
 	ret |= check_timerfd();
 	ret |= check_mnt_id();
+	ret |= check_aio_remap();
 
+out:
 	if (!ret)
-		pr_msg("Looks good.\n");
+		print_on_level(DEFAULT_LOGLEVEL, "Looks good.\n");
 
 	return ret;
 }
+
+static int check_tun(void)
+{
+	/*
+	 * In case there's no TUN support at all we
+	 * should report error. Unlike this plain criu
+	 * check would report "Looks good" in this case
+	 * since C/R effectively works, just not for TUN.
+	 */
+	return check_tun_cr(-1);
+}
+
+static int check_userns(void)
+{
+	int ret;
+	unsigned long size = 0;
+
+	ret = access("/proc/self/ns/user", F_OK);
+	if (ret) {
+		pr_perror("No userns proc file");
+		return -1;
+	}
+
+	ret = sys_prctl(PR_SET_MM, PR_SET_MM_MAP_SIZE, (unsigned long)&size, 0, 0);
+	if (ret) {
+		pr_perror("No new prctl API");
+		return -1;
+	}
+
+	return 0;
+}
+
+int check_add_feature(char *feat)
+{
+	if (!strcmp(feat, "mnt_id"))
+		chk_feature = check_mnt_id;
+	else if (!strcmp(feat, "aio_remap"))
+		chk_feature = check_aio_remap;
+	else if (!strcmp(feat, "timerfd"))
+		chk_feature = check_timerfd;
+	else if (!strcmp(feat, "tun"))
+		chk_feature = check_tun;
+	else if (!strcmp(feat, "userns"))
+		chk_feature = check_userns;
+	else {
+		pr_err("Unknown feature %s\n", feat);
+		return -1;
+	}
+
+	return 0;
+}
