@@ -11,7 +11,8 @@
 #define CPU_CAP_NONE		(0u)
 #define CPU_CAP_ALL		(-1u)
 #define CPU_CAP_FPU		(1u)		/* Only FPU capability required */
-#define CPU_CAP_CPU		(2u)		/* CPU capability required */
+#define CPU_CAP_CPU		(2u)		/* Strict CPU capability required */
+#define CPU_CAP_INS		(4u)		/* Instructions CPU capatibility */
 #define CPU_CAP_DEFAULT		(CPU_CAP_FPU)
 
 struct cg_root_opt {
@@ -43,6 +44,7 @@ struct cr_options {
 	struct list_head	veth_pairs;
 	struct list_head	scripts;
 	struct list_head	ext_mounts;
+	struct list_head	inherit_fds;
 	char			*libdir;
 	bool			use_page_server;
 	unsigned short		ps_port;
