@@ -5,24 +5,17 @@
 #include "asm/types.h"
 #include "image.h"
 #include "list.h"
+#include "cgroup.h"
 
 #include "protobuf/eventfd.pb-c.h"
 #include "protobuf/eventpoll.pb-c.h"
 #include "protobuf/signalfd.pb-c.h"
 #include "protobuf/fsnotify.pb-c.h"
+#include "protobuf/timerfd.pb-c.h"
 
 #define PROC_TASK_COMM_LEN	32
 #define PROC_TASK_COMM_LEN_FMT	"(%31s"
 
-struct proc_pid_stat_small {
-	int			pid;
-	char			comm[PROC_TASK_COMM_LEN];
-	char			state;
-	int			ppid;
-	int			pgid;
-	int			sid;
-};
-
 struct proc_pid_stat {
 	int			pid;
 	char			comm[PROC_TASK_COMM_LEN];
@@ -88,6 +81,9 @@ struct proc_status_creds {
 	u32 cap_prm[PROC_CAP_SIZE];
 	u32 cap_eff[PROC_CAP_SIZE];
 	u32 cap_bnd[PROC_CAP_SIZE];
+
+	char			state;
+	int			ppid;
 };
 
 struct mount_info;
@@ -99,22 +95,41 @@ struct fstype {
 	int (*parse)(struct mount_info *pm);
 };
 
+struct ext_mount;
 struct mount_info {
 	int		mnt_id;
 	int		parent_mnt_id;
 	unsigned int	s_dev;
 	char		*root;
+	/*
+	 * During dump mountpoint contains path with dot at the 
+	 * beginning. It allows to use openat, statat, etc without 
+	 * creating a temporary copy of the path.
+	 *
+	 * On restore mountpoint is prepended with so called ns
+	 * root path -- it's a place in fs where the namespace
+	 * mount tree is constructed. Check mnt_roots for details.
+	 * The ns_mountpoint contains path w/o this prefix.
+	 */
 	char		*mountpoint;
+	char		*ns_mountpoint;
 	unsigned	flags;
 	int		master_id;
 	int		shared_id;
 	struct fstype	*fstype;
 	char		*source;
 	char		*options;
-	bool		mounted;
+	union {
+		bool		mounted;
+		bool		dumped;
+	};
 	bool		need_plugin;
 	int		is_file;
+	bool		is_ns_root;
 	struct mount_info *next;
+	struct ns_id	*nsid;
+
+	struct ext_mount *external;
 
 	/* tree linkage */
 	struct mount_info *parent;
@@ -138,19 +153,50 @@ extern void mnt_entry_free(struct mount_info *mi);
 
 struct vm_area_list;
 
-extern struct mount_info *parse_mountinfo(pid_t pid);
+extern struct mount_info *parse_mountinfo(pid_t pid, struct ns_id *nsid);
 extern int parse_pid_stat(pid_t pid, struct proc_pid_stat *s);
-extern int parse_pid_stat_small(pid_t pid, struct proc_pid_stat_small *s);
-extern int parse_smaps(pid_t pid, struct vm_area_list *vma_area_list, bool use_map_files);
+extern int parse_smaps(pid_t pid, struct vm_area_list *vma_area_list);
 extern int parse_self_maps_lite(struct vm_area_list *vms);
 extern int parse_pid_status(pid_t pid, struct proc_status_creds *);
 
+struct inotify_wd_entry {
+	InotifyWdEntry e;
+	FhEntry f_handle;
+	struct list_head node;
+};
+
+struct fanotify_mark_entry {
+	FanotifyMarkEntry e;
+	FhEntry f_handle;
+	struct list_head node;
+	union {
+		FanotifyInodeMarkEntry ie;
+		FanotifyMountMarkEntry me;
+	};
+};
+
+struct eventpoll_tfd_entry {
+	EventpollTfdEntry e;
+	struct list_head node;
+};
+
 union fdinfo_entries {
 	EventfdFileEntry efd;
-	EventpollTfdEntry epl;
 	SignalfdEntry sfd;
-	InotifyWdEntry ify;
-	FanotifyMarkEntry ffy;
+	struct inotify_wd_entry ify;
+	struct fanotify_mark_entry ffy;
+	struct eventpoll_tfd_entry epl;
+	TimerfdEntry tfy;
+};
+
+extern void free_inotify_wd_entry(union fdinfo_entries *e);
+extern void free_fanotify_mark_entry(union fdinfo_entries *e);
+extern void free_event_poll_entry(union fdinfo_entries *e);
+
+struct fdinfo_common {
+	off64_t pos;
+	int flags;
+	int mnt_id;
 };
 
 extern int parse_fdinfo(int fd, int type,
@@ -159,8 +205,35 @@ extern int parse_fdinfo_pid(int pid, int fd, int type,
 		int (*cb)(union fdinfo_entries *e, void *arg), void *arg);
 extern int parse_cpuinfo_features(int (*handler)(char *tok));
 extern int parse_file_locks(void);
+extern int get_fd_mntid(int fd, int *mnt_id);
 
 struct pid;
 extern int parse_threads(int pid, struct pid **_t, int *_n);
 
+extern int check_mnt_id(void);
+
+/*
+ * This struct describes a group controlled by one controller.
+ * The @name is the controller name or 'name=...' for named cgroups.
+ * The @path is the path from the hierarchy root.
+ */
+
+struct cg_ctl {
+	struct list_head l;
+	char *name;
+	char *path;
+};
+
+/*
+ * Returns the list of cg_ctl-s sorted by name
+ */
+
+extern int parse_task_cgroup(int pid, struct list_head *l, unsigned int *n);
+extern void put_ctls(struct list_head *);
+
+int parse_cgroups(struct list_head *cgroups, unsigned int *n_cgroups);
+
+/* callback for AUFS support */
+extern int aufs_parse(struct mount_info *mi);
+
 #endif /* __CR_PROC_PARSE_H__ */
