@@ -2,13 +2,14 @@
 #include <unistd.h>
 #include <stdlib.h>
 
+#include "cr_options.h"
 #include "pstree.h"
-#include "restorer.h"
 #include "util.h"
 #include "lock.h"
 #include "namespaces.h"
 #include "files.h"
 #include "tty.h"
+#include "mount.h"
 #include "asm/dump.h"
 
 #include "protobuf.h"
@@ -18,61 +19,89 @@ struct pstree_item *root_item;
 
 void core_entry_free(CoreEntry *core)
 {
-	if (core) {
-		arch_free_thread_info(core);
-		if (core->thread_core)
-			xfree(core->thread_core->sas);
-		xfree(core->thread_core);
-		xfree(core->tc);
-		xfree(core->ids);
-	}
+	if (core->tc && core->tc->timers)
+		xfree(core->tc->timers->posix);
+	arch_free_thread_info(core);
+	xfree(core);
 }
 
-CoreEntry *core_entry_alloc(int alloc_thread_info, int alloc_tc)
+#ifndef RLIM_NLIMITS
+# define RLIM_NLIMITS 16
+#endif
+
+CoreEntry *core_entry_alloc(int th, int tsk)
 {
-	CoreEntry *core;
-	TaskCoreEntry *tc;
+	size_t sz;
+	CoreEntry *core = NULL;
+	void *m;
 
-	core = xmalloc(sizeof(*core));
-	if (!core)
-		return NULL;
-	core_entry__init(core);
-
-	core->mtype = CORE_ENTRY__MARCH;
-
-	if (alloc_thread_info) {
-		ThreadCoreEntry *thread_core;
-		ThreadSasEntry *sas;
-
-		if (arch_alloc_thread_info(core))
-			goto err;
-
-		thread_core = xmalloc(sizeof(*thread_core));
-		if (!thread_core)
-			goto err;
-		thread_core_entry__init(thread_core);
-		core->thread_core = thread_core;
-
-		sas = xmalloc(sizeof(*sas));
-		if (!sas)
-			goto err;
-		thread_sas_entry__init(sas);
-		core->thread_core->sas = sas;
+	sz = sizeof(CoreEntry);
+	if (tsk) {
+		sz += sizeof(TaskCoreEntry) + TASK_COMM_LEN;
+		if (th) {
+			sz += sizeof(TaskRlimitsEntry);
+			sz += RLIM_NLIMITS * sizeof(RlimitEntry *);
+			sz += RLIM_NLIMITS * sizeof(RlimitEntry);
+			sz += sizeof(TaskTimersEntry);
+			sz += 3 * sizeof(ItimerEntry); /* 3 for real, virt and prof */
+		}
 	}
+	if (th)
+		sz += sizeof(ThreadCoreEntry) + sizeof(ThreadSasEntry);
 
-	if (alloc_tc) {
-		tc = xzalloc(sizeof(*tc) + TASK_COMM_LEN);
-		if (!tc)
-			goto err;
-		task_core_entry__init(tc);
-		tc->comm = (void *)tc + sizeof(*tc);
-		core->tc = tc;
+	m = xmalloc(sz);
+	if (m) {
+		core = xptr_pull(&m, CoreEntry);
+		core_entry__init(core);
+		core->mtype = CORE_ENTRY__MARCH;
+
+		if (tsk) {
+			core->tc = xptr_pull(&m, TaskCoreEntry);
+			task_core_entry__init(core->tc);
+			core->tc->comm = xptr_pull_s(&m, TASK_COMM_LEN);
+			memzero(core->tc->comm, TASK_COMM_LEN);
+
+			if (th) {
+				TaskRlimitsEntry *rls;
+				TaskTimersEntry *tte;
+				int i;
+
+				rls = core->tc->rlimits = xptr_pull(&m, TaskRlimitsEntry);
+				task_rlimits_entry__init(rls);
+
+				rls->n_rlimits = RLIM_NLIMITS;
+				rls->rlimits = xptr_pull_s(&m, sizeof(RlimitEntry *) * RLIM_NLIMITS);
+
+				for (i = 0; i < RLIM_NLIMITS; i++) {
+					rls->rlimits[i] = xptr_pull(&m, RlimitEntry);
+					rlimit_entry__init(rls->rlimits[i]);
+				}
+
+				tte = core->tc->timers = xptr_pull(&m, TaskTimersEntry);
+				task_timers_entry__init(tte);
+				tte->real = xptr_pull(&m, ItimerEntry);
+				itimer_entry__init(tte->real);
+				tte->virt = xptr_pull(&m, ItimerEntry);
+				itimer_entry__init(tte->virt);
+				tte->prof = xptr_pull(&m, ItimerEntry);
+				itimer_entry__init(tte->prof);
+			}
+		}
+
+		if (th) {
+			core->thread_core = xptr_pull(&m, ThreadCoreEntry);
+			thread_core_entry__init(core->thread_core);
+			core->thread_core->sas = xptr_pull(&m, ThreadSasEntry);
+			thread_sas_entry__init(core->thread_core->sas);
+
+			if (arch_alloc_thread_info(core)) {
+				xfree(core);
+				core = NULL;
+			}
+		}
 	}
 
 	return core;
-err:
-	core_entry_free(core);
-	return NULL;
 }
 
 int pstree_alloc_cores(struct pstree_item *item)
@@ -133,10 +162,22 @@ void free_pstree(struct pstree_item *root_item)
 struct pstree_item *__alloc_pstree_item(bool rst)
 {
 	struct pstree_item *item;
+	int sz;
 
-	item = xzalloc(sizeof(*item) + (rst ? sizeof(item->rst[0]) : 0));
-	if (!item)
-		return NULL;
+	if (!rst) {
+		sz = sizeof(*item) + sizeof(struct dmp_info);
+		item = xzalloc(sz);
+		if (!item)
+			return NULL;
+	} else {
+		sz = sizeof(*item) + sizeof(struct rst_info);
+		item = shmalloc(sz);
+		if (!item)
+			return NULL;
+
+		memset(item, 0, sz);
+		vm_area_list_init(&rsti(item)->vmas);
+	}
 
 	INIT_LIST_HEAD(&item->children);
 	INIT_LIST_HEAD(&item->sibling);
@@ -148,6 +189,20 @@ struct pstree_item *__alloc_pstree_item(bool rst)
 	return item;
 }
 
+struct pstree_item *alloc_pstree_helper(void)
+{
+	struct pstree_item *ret;
+
+	ret = alloc_pstree_item_with_rst();
+	if (ret) {
+		ret->state = TASK_HELPER;
+		rsti(ret)->clone_flags = CLONE_FILES | CLONE_FS;
+		task_entries->nr_helpers++;
+	}
+
+	return ret;
+}
+
 /* Deep first search on children */
 struct pstree_item *pstree_item_next(struct pstree_item *item)
 {
@@ -168,7 +223,7 @@ int dump_pstree(struct pstree_item *root_item)
 	struct pstree_item *item = root_item;
 	PstreeEntry e = PSTREE_ENTRY__INIT;
 	int ret = -1, i;
-	int pstree_fd;
+	struct cr_img *img;
 
 	pr_info("\n");
 	pr_info("Dumping pstree (pid: %d)\n", root_item->pid.real);
@@ -191,8 +246,8 @@ int dump_pstree(struct pstree_item *root_item)
 		}
 	}
 
-	pstree_fd = open_image(CR_FD_PSTREE, O_DUMP);
-	if (pstree_fd < 0)
+	img = open_image(CR_FD_PSTREE, O_DUMP);
+	if (!img)
 		return -1;
 
 	for_each_pstree_item(item) {
@@ -211,7 +266,7 @@ int dump_pstree(struct pstree_item *root_item)
 		for (i = 0; i < item->nr_threads; i++)
 			e.threads[i] = item->threads[i].virt;
 
-		ret = pb_write_one(pstree_fd, &e, PB_PSTREE);
+		ret = pb_write_one(img, &e, PB_PSTREE);
 		xfree(e.threads);
 
 		if (ret)
@@ -221,7 +276,7 @@ int dump_pstree(struct pstree_item *root_item)
 
 err:
 	pr_info("----------------------------------------\n");
-	close(pstree_fd);
+	close_image(img);
 	return ret;
 }
 
@@ -281,19 +336,20 @@ static int prepare_pstree_for_shell_job(void)
 
 static int read_pstree_image(void)
 {
-	int ret = 0, i, ps_fd, fd;
+	int ret = 0, i;
+	struct cr_img *img;
 	struct pstree_item *pi, *parent = NULL;
 
 	pr_info("Reading image tree\n");
 
-	ps_fd = open_image(CR_FD_PSTREE, O_RSTR);
-	if (ps_fd < 0)
-		return ps_fd;
+	img = open_image(CR_FD_PSTREE, O_RSTR);
+	if (!img)
+		return -1;
 
 	while (1) {
 		PstreeEntry *e;
 
-		ret = pb_read_one_eof(ps_fd, &e, PB_PSTREE);
+		ret = pb_read_one_eof(img, &e, PB_PSTREE);
 		if (ret <= 0)
 			break;
 
@@ -356,28 +412,40 @@ static int read_pstree_image(void)
 		if (!pi->threads)
 			break;
 
-		for (i = 0; i < e->n_threads; i++)
+		for (i = 0; i < e->n_threads; i++) {
+			pi->threads[i].real = -1;
 			pi->threads[i].virt = e->threads[i];
+			max_pid = max((int)e->threads[i], max_pid);
+		}
 
 		task_entries->nr_threads += e->n_threads;
 		task_entries->nr_tasks++;
 
 		pstree_entry__free_unpacked(e, NULL);
 
-		fd = open_image(CR_FD_IDS, O_RSTR, pi->pid.virt);
-		if (fd < 0) {
-			if (errno == ENOENT)
-				continue;
-			goto err;
-		}
-		ret = pb_read_one(fd, &pi->ids, PB_IDS);
-		close(fd);
+		{
+			struct cr_img *img;
+
+			img = open_image(CR_FD_IDS, O_RSTR, pi->pid.virt);
+			if (!img) {
+				if (errno == ENOENT)
+					continue;
+				goto err;
+			}
+			ret = pb_read_one(img, &pi->ids, PB_IDS);
+			close_image(img);
+			}
+
 		if (ret != 1)
 			goto err;
 
+		if (pi->ids->has_mnt_ns_id) {
+			if (rst_add_ns_id(pi->ids->mnt_ns_id, pi->pid.virt, &mnt_ns_desc))
+				goto err;
+		}
 	}
 err:
-	close(ps_fd);
+	close_image(img);
 	return ret;
 }
 
@@ -404,16 +472,14 @@ static int prepare_pstree_ids(void)
 		if (item->sid == root_item->sid || item->sid == item->pid.virt)
 			continue;
 
-		helper = alloc_pstree_item_with_rst();
+		helper = alloc_pstree_helper();
 		if (helper == NULL)
 			return -1;
 		helper->sid = item->sid;
 		helper->pgid = item->sid;
 		helper->pid.virt = item->sid;
-		helper->state = TASK_HELPER;
 		helper->parent = root_item;
 		list_add_tail(&helper->sibling, &helpers);
-		task_entries->nr_helpers++;
 
 		pr_info("Add a helper %d for restoring SID %d\n",
 				helper->pid.virt, helper->sid);
@@ -511,8 +577,10 @@ static int prepare_pstree_ids(void)
 				break;
 		}
 
-		if (gleader)
+		if (gleader) {
+			rsti(item)->pgrp_leader = gleader;
 			continue;
+		}
 
 		/*
 		 * If the PGID is eq to current one -- this
@@ -522,16 +590,15 @@ static int prepare_pstree_ids(void)
 		if (current_pgid == item->pgid)
 			continue;
 
-		helper = alloc_pstree_item_with_rst();
+		helper = alloc_pstree_helper();
 		if (helper == NULL)
 			return -1;
 		helper->sid = item->sid;
 		helper->pgid = item->pgid;
 		helper->pid.virt = item->pgid;
-		helper->state = TASK_HELPER;
 		helper->parent = item;
 		list_add(&helper->sibling, &item->children);
-		task_entries->nr_helpers++;
+		rsti(item)->pgrp_leader = helper;
 
 		pr_info("Add a helper %d for restoring PGID %d\n",
 				helper->pid.virt, helper->pgid);
@@ -557,6 +624,8 @@ static unsigned long get_clone_mask(TaskKobjIdsEntry *i,
 		mask |= CLONE_NEWUTS;
 	if (i->mnt_ns_id != p->mnt_ns_id)
 		mask |= CLONE_NEWNS;
+	if (i->user_ns_id != p->user_ns_id)
+		mask |= CLONE_NEWUSER;
 
 	return mask;
 }
@@ -615,24 +684,38 @@ static int prepare_pstree_kobj_ids(void)
 		}
 
 set_mask:
-		item->rst->clone_flags = cflags;
+		rsti(item)->clone_flags = cflags;
+		if (parent)
+			/*
+			 * Mount namespaces are setns()-ed at
+			 * restore_task_mnt_ns() explicitly,
+			 * no need in creating it with its own
+			 * temporary namespace
+			 */
+			rsti(item)->clone_flags &= ~CLONE_NEWNS;
+
+		cflags &= CLONE_ALLNS;
 
-		/*
-		 * Workaround for current namespaces model --
-		 * all tasks should be in one namespace. And
-		 * this namespace is either inherited from the
-		 * criu or is created for the init task (only)
-		 */
 		if (item == root_item) {
 			pr_info("Will restore in %lx namespaces\n", cflags);
-			current_ns_mask = cflags & CLONE_ALLNS;
-		} else if (cflags & CLONE_ALLNS) {
+			root_ns_mask = cflags;
+		} else if (cflags & ~(root_ns_mask & CLONE_SUBNS)) {
+			/*
+			 * Namespaces from CLONE_SUBNS can be nested, but in
+			 * this case nobody can't share external namespaces of
+			 * these types.
+			 *
+			 * Workaround for all other namespaces --
+			 * all tasks should be in one namespace. And
+			 * this namespace is either inherited from the
+			 * criu or is created for the init task (only)
+			 */
 			pr_err("Can't restore sub-task in NS\n");
 			return -1;
 		}
 	}
 
-	pr_debug("NS mask to use %lx\n", current_ns_mask);
+	pr_debug("NS mask to use %lx\n", root_ns_mask);
 	return 0;
 }
 
