@@ -21,10 +21,6 @@
 #include "net.h"
 #include "fs-magic.h"
 
-#ifndef NETLINK_SOCK_DIAG
-#define NETLINK_SOCK_DIAG NETLINK_INET_DIAG
-#endif
-
 #ifndef SOCK_DIAG_BY_FAMILY
 #define SOCK_DIAG_BY_FAMILY 20
 #endif
@@ -35,6 +31,22 @@
 #define SO_GET_FILTER	SO_ATTACH_FILTER
 #endif
 
+struct sock_diag_greq {
+	u8	family;
+	u8	protocol;
+};
+
+struct sock_diag_req {
+	struct nlmsghdr hdr;
+	union {
+		struct unix_diag_req	u;
+		struct inet_diag_req_v2	i;
+		struct packet_diag_req	p;
+		struct netlink_diag_req n;
+		struct sock_diag_greq	g;
+	} r;
+};
+
 enum socket_cl_bits
 {
 	NETLINK_CL_BIT,
@@ -100,6 +112,95 @@ bool socket_test_collect_bit(unsigned int family, unsigned int proto)
 	return test_bit(nr, socket_cl_bits) != 0;
 }
 
+static int probe_recv_one(struct nlmsghdr *h, void *arg)
+{
+	pr_err("PROBE RECEIVED\n");
+	return -1;
+}
+
+static int probe_err(int err, void *arg)
+{
+	int expected_err = *(int *)arg;
+
+	if (err == expected_err)
+		return 0;
+
+	pr_err("Diag module missing (%d)\n", err);
+	return err;
+}
+
+static inline void probe_diag(int nl, struct sock_diag_req *req, int expected_err)
+{
+	do_rtnl_req(nl, req, req->hdr.nlmsg_len, probe_recv_one, probe_err, &expected_err);
+}
+
+void preload_socket_modules()
+{
+	int nl;
+	struct sock_diag_req req;
+
+	/*
+	 * If the task to dump (e.g. an LXC container) has any netlink
+	 * KOBJECT_UEVENT socket open and the _diag modules aren't
+	 * loaded is dumped, criu will freeze the task and then the
+	 * kernel will send it messages on the socket, and then we will
+	 * fail to dump because the socket has pending data. The Real
+	 * Solution is to dump this pending data, but we just make sure
+	 * modules are there beforehand for now so that the first dump
+	 * doesn't fail.
+	 */
+
+	nl = socket(PF_NETLINK, SOCK_RAW, NETLINK_SOCK_DIAG);
+	if (nl < 0)
+		return;
+
+	pr_info("Probing sock diag modules\n");
+
+	memset(&req, 0, sizeof(req));
+	req.hdr.nlmsg_type	= SOCK_DIAG_BY_FAMILY;
+	req.hdr.nlmsg_seq	= CR_NLMSG_SEQ;
+
+	/*
+	 * Probe UNIX, netlink and packet diag-s by feeding
+	 * to the kernel request that is shorter than they
+	 * expect, byt still containing the family to make
+	 * sure the family handler is there. The family-level
+	 * diag module would report EINVAL in this case.
+	 */
+
+	req.hdr.nlmsg_len = sizeof(req.hdr) + sizeof(req.r.g);
+	req.hdr.nlmsg_flags = NLM_F_DUMP | NLM_F_REQUEST;
+
+	req.r.g.family = AF_UNIX;
+	probe_diag(nl, &req, -EINVAL);
+
+	req.r.g.family = AF_PACKET;
+	probe_diag(nl, &req, -EINVAL);
+
+	req.r.g.family = AF_NETLINK;
+	probe_diag(nl, &req, -EINVAL);
+
+	/*
+	 * TCP and UDP(LITE) diags do not support such trick, only
+	 * inet_diag module can be probed like that. For the protocol
+	 * level ones it's OK to request for exact non-existing socket
+	 * and check for ENOENT being reported back as error.
+	 */
+
+	req.hdr.nlmsg_len = sizeof(req.hdr) + sizeof(req.r.i);
+	req.hdr.nlmsg_flags = NLM_F_REQUEST;
+	req.r.i.sdiag_family = AF_INET;
+
+	req.r.i.sdiag_protocol = IPPROTO_TCP;
+	probe_diag(nl, &req, -ENOENT);
+
+	req.r.i.sdiag_protocol = IPPROTO_UDP; /* UDLITE is merged with UDP */
+	probe_diag(nl, &req, -ENOENT);
+
+	close(nl);
+	pr_info("Done probing\n");
+}
+
 static int dump_bound_dev(int sk, SkOptsEntry *soe)
 {
 	int ret;
@@ -273,6 +374,18 @@ int do_restore_opt(int sk, int level, int name, void *val, int len)
 	return 0;
 }
 
+static int sk_setbufs(void *arg, int fd)
+{
+	u32 *buf = (u32 *)arg;
+
+	if (restore_opt(fd, SOL_SOCKET, SO_SNDBUFFORCE, &buf[0]))
+		return -1;
+	if (restore_opt(fd, SOL_SOCKET, SO_RCVBUFFORCE, &buf[1]))
+		return -1;
+
+	return 0;
+}
+
 /*
  * Set sizes of buffers to maximum and prevent blocking
  * Caller of this fn should call other socket restoring
@@ -282,13 +395,10 @@ int do_restore_opt(int sk, int level, int name, void *val, int len)
 int restore_prepare_socket(int sk)
 {
 	int flags;
-
 	/* In kernel a bufsize has type int and a value is doubled. */
-	u32 maxbuf = INT_MAX / 2;
+	u32 maxbuf[2] = { INT_MAX / 2, INT_MAX / 2 };
 
-	if (restore_opt(sk, SOL_SOCKET, SO_SNDBUFFORCE, &maxbuf))
-		return -1;
-	if (restore_opt(sk, SOL_SOCKET, SO_RCVBUFFORCE, &maxbuf))
+	if (userns_call(sk_setbufs, 0, maxbuf, sizeof(maxbuf), sk))
 		return -1;
 
 	/* Prevent blocking on restore */
@@ -309,14 +419,12 @@ int restore_socket_opts(int sk, SkOptsEntry *soe)
 {
 	int ret = 0, val;
 	struct timeval tv;
+	u32 bufs[2] = { soe->so_sndbuf, soe->so_rcvbuf };
 
 	pr_info("%d restore sndbuf %d rcv buf %d\n", sk, soe->so_sndbuf, soe->so_rcvbuf);
 
 	/* setsockopt() multiplies the input values by 2 */
-	val = soe->so_sndbuf / 2;
-	ret |= restore_opt(sk, SOL_SOCKET, SO_SNDBUFFORCE, &val);
-	val = soe->so_rcvbuf / 2;
-	ret |= restore_opt(sk, SOL_SOCKET, SO_RCVBUFFORCE, &val);
+	ret |= userns_call(sk_setbufs, UNS_ASYNC, bufs, sizeof(bufs), sk);
 
 	if (soe->has_so_priority) {
 		pr_debug("\trestore priority %d for socket\n", soe->so_priority);
@@ -439,7 +547,7 @@ void release_skopts(SkOptsEntry *soe)
 	xfree(soe->so_bound_dev);
 }
 
-int dump_socket(struct fd_parms *p, int lfd, const int fdinfo)
+int dump_socket(struct fd_parms *p, int lfd, struct cr_img *img)
 {
 	int family;
 	const struct fdtype_ops *ops;
@@ -468,7 +576,7 @@ int dump_socket(struct fd_parms *p, int lfd, const int fdinfo)
 		return -1;
 	}
 
-	return do_dump_gen_file(p, lfd, ops, fdinfo);
+	return do_dump_gen_file(p, lfd, ops, img);
 }
 
 static int inet_receive_one(struct nlmsghdr *h, void *arg)
@@ -492,22 +600,12 @@ static int inet_receive_one(struct nlmsghdr *h, void *arg)
 	return inet_collect_one(h, i->sdiag_family, type);
 }
 
-struct sock_diag_req {
-	struct nlmsghdr hdr;
-	union {
-		struct unix_diag_req	u;
-		struct inet_diag_req_v2	i;
-		struct packet_diag_req	p;
-		struct netlink_diag_req n;
-	} r;
-};
-
 static int do_collect_req(int nl, struct sock_diag_req *req, int size,
 		int (*receive_callback)(struct nlmsghdr *h, void *), void *arg)
 {
 	int tmp;
 
-	tmp = do_rtnl_req(nl, req, size, receive_callback, arg);
+	tmp = do_rtnl_req(nl, req, size, receive_callback, NULL, arg);
 
 	if (tmp == 0)
 		set_collect_bit(req->r.n.sdiag_family, req->r.n.sdiag_protocol);
@@ -515,27 +613,12 @@ static int do_collect_req(int nl, struct sock_diag_req *req, int size,
 	return tmp;
 }
 
-int collect_sockets(int pid)
+int collect_sockets(struct ns_id *ns)
 {
 	int err = 0, tmp;
-	int rst = -1;
-	int nl;
+	int nl = ns->net.nlsk;
 	struct sock_diag_req req;
 
-	if (root_ns_mask & CLONE_NEWNET) {
-		pr_info("Switching to %d's net for collecting sockets\n", pid);
-
-		if (switch_ns(pid, &net_ns_desc, &rst))
-			return -1;
-	}
-
-	nl = socket(PF_NETLINK, SOCK_RAW, NETLINK_SOCK_DIAG);
-	if (nl < 0) {
-		pr_perror("Can't create sock diag socket");
-		err = -1;
-		goto out;
-	}
-
 	memset(&req, 0, sizeof(req));
 	req.hdr.nlmsg_len	= sizeof(req);
 	req.hdr.nlmsg_type	= SOCK_DIAG_BY_FAMILY;
@@ -615,7 +698,7 @@ int collect_sockets(int pid)
 	tmp = do_collect_req(nl, &req, sizeof(req), packet_receive_one, NULL);
 	if (tmp) {
 		pr_warn("The current kernel doesn't support packet_diag\n");
-		if (pid == 0 || tmp != -ENOENT) /* Fedora 19 */
+		if (ns->pid == 0 || tmp != -ENOENT) /* Fedora 19 */
 			err = tmp;
 	}
 
@@ -625,16 +708,15 @@ int collect_sockets(int pid)
 	tmp = do_collect_req(nl, &req, sizeof(req), netlink_receive_one, NULL);
 	if (tmp) {
 		pr_warn("The current kernel doesn't support netlink_diag\n");
-		if (pid == 0 || tmp != -ENOENT) /* Fedora 19 */
+		if (ns->pid == 0 || tmp != -ENOENT) /* Fedora 19 */
 			err = tmp;
 	}
 
+	/* don't need anymore */
 	close(nl);
-out:
-	if (rst >= 0) {
-		if (restore_ns(rst, &net_ns_desc) < 0)
-			err = -1;
-	} else if (pid != 0) {
+	ns->net.nlsk = -1;
+
+	if (ns->pid == getpid()) {
 		/*
 		 * If netns isn't dumped, criu will fail only
 		 * if an unsupported socket will be really dumped.
